#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include <X11/Xatom.h>
#include <X11/extensions/Xrender.h>

#include <compiz.h>


#define BS_SATURATION_INCREASE_BUTTON_DEFAULT       Button4
#define BS_SATURATION_INCREASE_MODIFIERS_DEFAULT    ControlMask
#define BS_SATURATION_DECREASE_BUTTON_DEFAULT       Button5
#define BS_SATURATION_DECREASE_MODIFIERS_DEFAULT    ControlMask


#define BS_BRIGHTNESS_INCREASE_BUTTON_DEFAULT       Button4
#define BS_BRIGHTNESS_INCREASE_MODIFIERS_DEFAULT    ShiftMask
#define BS_BRIGHTNESS_DECREASE_BUTTON_DEFAULT       Button5
#define BS_BRIGHTNESS_DECREASE_MODIFIERS_DEFAULT    ShiftMask


#define BS_SATURATION_STEP_DEFAULT                  5
#define BS_SATURATION_STEP_MAX                      10
#define BS_SATURATION_STEP_MIN                      1

#define BS_BRIGHTNESS_STEP_DEFAULT                  5
#define BS_BRIGHTNESS_STEP_MAX                      10
#define BS_BRIGHTNESS_STEP_MIN                      1


static int displayPrivateIndex;

typedef struct _BSDisplay {
    int		    screenPrivateIndex;
    HandleEventProc handleEvent;
} BSDisplay;



#define BS_SCREEN_OPTION_SATURATION_INCREASE    0
#define BS_SCREEN_OPTION_SATURATION_DECREASE    1
#define BS_SCREEN_OPTION_BRIGHTNESS_INCREASE    2
#define BS_SCREEN_OPTION_BRIGHTNESS_DECREASE    3
#define BS_SCREEN_OPTION_SATURATION_STEP        4
#define BS_SCREEN_OPTION_BRIGHTNESS_STEP        5
#define BS_SCREEN_OPTION_NUM                    6

typedef struct _BSSCreen {
    CompOption opt[BS_SCREEN_OPTION_NUM];
    int brightnessStep;
    int saturationStep;
} BSScreen;


#define GET_BS_DISPLAY(d) ((BSDisplay *) (d)->privates[displayPrivateIndex].ptr)
#define BS_DISPLAY(d) BSDisplay *bd = GET_BS_DISPLAY (d)
#define GET_BS_SCREEN(s, bd) ((BSScreen *) (s)->privates[(bd)->screenPrivateIndex].ptr)
#define BS_SCREEN(s) BSScreen *bs = GET_BS_SCREEN (s, GET_BS_DISPLAY (s->display))


#define NUM_OPTIONS(s) (sizeof ((s)->opt) / sizeof (CompOption))

static CompOption *
BSGetScreenOptions (CompScreen *screen, int *count)
{
    BS_SCREEN (screen);

    *count = NUM_OPTIONS (bs);
    return bs->opt;
}

static Bool
BSSetScreenOption (CompScreen      *screen,
		       char	       *name,
		       CompOptionValue *value)
{
    CompOption *o;
    int	       index;

    BS_SCREEN (screen);

    o = compFindOption (bs->opt, NUM_OPTIONS (bs), name, &index);
    if (!o)
	return FALSE;

    switch (index) {
        case BS_SCREEN_OPTION_SATURATION_INCREASE:
        case BS_SCREEN_OPTION_SATURATION_DECREASE:
        case BS_SCREEN_OPTION_BRIGHTNESS_INCREASE:
        case BS_SCREEN_OPTION_BRIGHTNESS_DECREASE:
            if (addScreenBinding (screen, &value->bind))
	    {
	        removeScreenBinding (screen, &o->value.bind);

	        if (compSetBindingOption (o, value))
		    return TRUE;
	    }
	    break;

        case BS_SCREEN_OPTION_BRIGHTNESS_STEP:
            if (compSetIntOption (o, value))
            {
                bs->brightnessStep = o->value.i;
                return TRUE;
            }
            break;

        case BS_SCREEN_OPTION_SATURATION_STEP:
            if (compSetIntOption (o, value))
            {
                bs->saturationStep = o->value.i;
                return TRUE;
            }
            break;

        default:
            break;
    }
    return FALSE;
}

static void
BSScreenInitOptions (BSScreen *bs, Display *display)
{
    CompOption *o;

    o = &bs->opt[BS_SCREEN_OPTION_SATURATION_INCREASE];
    o->name                       = "saturation_increase";
    o->shortDesc                  = "Increase Saturation";
    o->longDesc                   = "Increase Saturation";
    o->type                       = CompOptionTypeBinding;
    o->value.bind.type		  = CompBindingTypeButton;
    o->value.bind.u.button.modifiers = BS_SATURATION_INCREASE_MODIFIERS_DEFAULT;
    o->value.bind.u.button.button    = BS_SATURATION_INCREASE_BUTTON_DEFAULT;



    o = &bs->opt[BS_SCREEN_OPTION_SATURATION_DECREASE];
    o->name                       = "saturation_decrease";
    o->shortDesc                  = "Decrease Saturation";
    o->longDesc                   = "Decrease Saturation";
    o->type                       = CompOptionTypeBinding;
    o->value.bind.type		  = CompBindingTypeButton;
    o->value.bind.u.button.modifiers = BS_SATURATION_DECREASE_MODIFIERS_DEFAULT;
    o->value.bind.u.button.button    = BS_SATURATION_DECREASE_BUTTON_DEFAULT;



    o = &bs->opt[BS_SCREEN_OPTION_BRIGHTNESS_INCREASE];
    o->name                       = "brightness_increase";
    o->shortDesc                  = "Increase Brightness";
    o->longDesc                   = "Increase Brightness";
    o->type                       = CompOptionTypeBinding;
    o->value.bind.type		  = CompBindingTypeButton;
    o->value.bind.u.button.modifiers = BS_BRIGHTNESS_INCREASE_MODIFIERS_DEFAULT;
    o->value.bind.u.button.button    = BS_BRIGHTNESS_INCREASE_BUTTON_DEFAULT;



    o = &bs->opt[BS_SCREEN_OPTION_BRIGHTNESS_DECREASE];
    o->name                       = "brightness_decrease";
    o->shortDesc                  = "Decrease Brightness";
    o->longDesc                   = "Decrease Brightness";
    o->type                       = CompOptionTypeBinding;
    o->value.bind.type		  = CompBindingTypeButton;
    o->value.bind.u.button.modifiers = BS_BRIGHTNESS_DECREASE_MODIFIERS_DEFAULT;
    o->value.bind.u.button.button    = BS_BRIGHTNESS_DECREASE_BUTTON_DEFAULT;



    o = &bs->opt[BS_SCREEN_OPTION_BRIGHTNESS_STEP];
    o->name		= "brightness_step";
    o->shortDesc	= "Brightness Step";
    o->longDesc		= "Brightness change step";
    o->type		= CompOptionTypeInt;
    o->value.i		= BS_BRIGHTNESS_STEP_DEFAULT;
    o->rest.i.min	= BS_BRIGHTNESS_STEP_MIN;
    o->rest.i.max	= BS_BRIGHTNESS_STEP_MAX;



    o = &bs->opt[BS_SCREEN_OPTION_SATURATION_STEP];
    o->name		= "saturation_step";
    o->shortDesc	= "Saturation Step";
    o->longDesc		= "Saturation change step";
    o->type		= CompOptionTypeInt;
    o->value.i		= BS_SATURATION_STEP_DEFAULT;
    o->rest.i.min	= BS_SATURATION_STEP_MIN;
    o->rest.i.max	= BS_SATURATION_STEP_MAX;

}

static void
changeWindowSaturation (CompWindow *w, int direction)
{
    int step, saturation;

    BS_SCREEN (w->screen);

    step = (COLOR * bs->saturationStep / 100);         

    saturation = w->paint.saturation + step * direction;
    if (saturation > COLOR)
    {
	saturation = COLOR;
    }
    else if (saturation < 0)
    {
	saturation = 0;
    }

    if (w->paint.saturation != saturation)
    {
	w->paint.saturation = saturation;

	setWindowProp32 (w->screen->display, w->id,
			 w->screen->display->winSaturationAtom,
			 w->paint.saturation);
	addWindowDamage (w);
    }
}

static void
changeWindowBrightness (CompWindow *w, int direction)
{
    int step, brightness;

    BS_SCREEN (w->screen);

    step = (BRIGHT * bs->brightnessStep / 100);   

    brightness = w->paint.brightness + step * direction;
    if (brightness > BRIGHT)
    {
	brightness = BRIGHT;
    }
    else if (brightness < 0)
    {
	brightness = 0;
    }

    if (w->paint.brightness != brightness)
    {
	w->paint.brightness = brightness;

	setWindowProp32 (w->screen->display, w->id,
			 w->screen->display->winBrightnessAtom,
			 w->paint.brightness);
	addWindowDamage (w);
    }
}


static void
BSHandleEvent (CompDisplay *d, XEvent *event)
{
    CompScreen *s;
    CompWindow *w;

    BS_DISPLAY (d);

    switch (event->type) {
        case KeyPress:
        case KeyRelease:
            s = findScreenAtDisplay (d, event->xkey.root);
            if (s)
            {
                BS_SCREEN (s);

                if (eventMatches (d, event, &bs->opt[BS_SCREEN_OPTION_SATURATION_INCREASE])) {
                    w = findTopLevelWindowAtScreen (s, event->xkey.window);
                    if (w) changeWindowSaturation (w, 1);
                }

                if (eventMatches (d, event, &bs->opt[BS_SCREEN_OPTION_SATURATION_DECREASE])) {
                    w = findTopLevelWindowAtScreen (s, event->xkey.window);
                    if (w) changeWindowSaturation (w, -1);
                }

                if (eventMatches (d, event, &bs->opt[BS_SCREEN_OPTION_BRIGHTNESS_INCREASE])) {
                    w = findTopLevelWindowAtScreen (s, event->xkey.window);
                    if (w) changeWindowBrightness (w, 1);
                }

                if (eventMatches (d, event, &bs->opt[BS_SCREEN_OPTION_BRIGHTNESS_DECREASE])) {
                    w = findTopLevelWindowAtScreen (s, event->xkey.window);
                    if (w) changeWindowBrightness (w, -1);
                }

            }

        case ButtonPress:
        case ButtonRelease:
            s = findScreenAtDisplay (d, event->xkey.root);
            if (s)
            {
                BS_SCREEN (s);

                if (eventMatches (d, event, &bs->opt[BS_SCREEN_OPTION_SATURATION_INCREASE])) {
                    w = findTopLevelWindowAtScreen (s, event->xbutton.window);
                    if (w) changeWindowSaturation (w, 1);
                }

                if (eventMatches (d, event, &bs->opt[BS_SCREEN_OPTION_SATURATION_DECREASE])) {
                    w = findTopLevelWindowAtScreen (s, event->xbutton.window);
                    if (w) changeWindowSaturation (w, -1);
                }

                if (eventMatches (d, event, &bs->opt[BS_SCREEN_OPTION_BRIGHTNESS_INCREASE])) {
                    w = findTopLevelWindowAtScreen (s, event->xbutton.window);
                    if (w) changeWindowBrightness (w, 1);
                }

                if (eventMatches (d, event, &bs->opt[BS_SCREEN_OPTION_BRIGHTNESS_DECREASE])) {
                    w = findTopLevelWindowAtScreen (s, event->xbutton.window);
                    if (w) changeWindowBrightness (w, -1);
                }

            }
    }

    UNWRAP (bd, d, handleEvent);
    (*d->handleEvent) (d, event);
    WRAP (bd, d, handleEvent, BSHandleEvent);
}

static Bool
BSInitDisplay (CompPlugin  *p, CompDisplay *d)
{
    BSDisplay *bd;

    bd = malloc (sizeof (BSDisplay));
    if (!bd)
	return FALSE;

    bd->screenPrivateIndex = allocateScreenPrivateIndex (d);
    if (bd->screenPrivateIndex < 0)
    {
	free (bd);
	return FALSE;
    }

    WRAP (bd, d, handleEvent, BSHandleEvent);

    d->privates[displayPrivateIndex].ptr = bd;

    return TRUE;
}

static void
BSFiniDisplay (CompPlugin  *p, CompDisplay *d)
{
    BS_DISPLAY (d);

    freeScreenPrivateIndex (d, bd->screenPrivateIndex);

    UNWRAP (bd, d, handleEvent);

    free (bd);
}

static Bool
BSInitScreen (CompPlugin *p, CompScreen *s)
{
    BSScreen *bs;
    BS_DISPLAY (s->display);

    bs = malloc (sizeof (BSScreen));
    if (!bs)
	return FALSE;

    bs->saturationStep = BS_SATURATION_STEP_DEFAULT;
    bs->brightnessStep = BS_BRIGHTNESS_STEP_DEFAULT;

    BSScreenInitOptions (bs, s->display->display);

    addScreenBinding (s, &bs->opt[BS_SCREEN_OPTION_BRIGHTNESS_INCREASE].value.bind);
    addScreenBinding (s, &bs->opt[BS_SCREEN_OPTION_BRIGHTNESS_DECREASE].value.bind);
    addScreenBinding (s, &bs->opt[BS_SCREEN_OPTION_SATURATION_INCREASE].value.bind);
    addScreenBinding (s, &bs->opt[BS_SCREEN_OPTION_SATURATION_DECREASE].value.bind);
    
    s->privates[bd->screenPrivateIndex].ptr = bs;

    return TRUE;
}


static void
BSFiniScreen (CompPlugin *p, CompScreen *s)
{
    BS_SCREEN (s);
    free (bs);
}

static Bool
BSInit (CompPlugin *p)
{
    displayPrivateIndex = allocateDisplayPrivateIndex ();
    if (displayPrivateIndex < 0)
	return FALSE;

    return TRUE;
}

static void
BSFini (CompPlugin *p)
{
    if (displayPrivateIndex >= 0)
	freeDisplayPrivateIndex (displayPrivateIndex);
}

CompPluginVTable BSVTable = {
    "bs",
    "Brightness and Saturation",
    "Brightness and Saturation adjustments",
    BSInit,
    BSFini,
    BSInitDisplay,
    BSFiniDisplay,
    BSInitScreen,
    BSFiniScreen,
    0, /* InitWindow */
    0, /* FiniWindow */
    0, /* GetDisplayOptions */
    0, /* SetDisplayOption */
    BSGetScreenOptions,
    BSSetScreenOption,
    0,
    0
};

CompPluginVTable *
getCompPluginInfo (void)
{
    return &BSVTable;
}


